package ru.dbotthepony.mc.otm.block.entity;

import com.google.common.collect.ImmutableList;
import net.minecraft.core.BlockPos;
import net.minecraft.core.Direction;
import net.minecraft.nbt.CompoundTag;
import net.minecraft.network.chat.Component;
import net.minecraft.network.chat.TranslatableComponent;
import net.minecraft.world.entity.player.Inventory;
import net.minecraft.world.entity.player.Player;
import net.minecraft.world.inventory.AbstractContainerMenu;
import net.minecraft.world.item.ItemStack;
import net.minecraft.world.level.Level;
import net.minecraft.world.level.block.state.BlockState;
import net.minecraftforge.common.capabilities.Capability;
import net.minecraftforge.common.util.LazyOptional;
import net.minecraftforge.items.CapabilityItemHandler;
import net.minecraftforge.items.IItemHandler;
import ru.dbotthepony.mc.otm.IMatterGridCell;
import ru.dbotthepony.mc.otm.MatterGrid;
import ru.dbotthepony.mc.otm.Registry;
import ru.dbotthepony.mc.otm.capability.IMatterHandler;
import ru.dbotthepony.mc.otm.capability.IPatternStorage;
import ru.dbotthepony.mc.otm.capability.MatteryCapability;
import ru.dbotthepony.mc.otm.container.MatteryContainer;
import ru.dbotthepony.mc.otm.menu.PatternStorageMenu;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.concurrent.atomic.AtomicInteger;

public class BlockEntityPatternStorage extends BlockEntityMattery implements IMatterGridCell, IItemHandler, IPatternStorage {
	public BlockEntityPatternStorage(BlockPos p_155229_, BlockState p_155230_) {
		super(Registry.BlockEntities.PATTERN_STORAGE, p_155229_, p_155230_);
	}

	public MatteryContainer patterns = new MatteryContainer(this::setChangedPatterns, 3 * 3);

	private void setChangedPatterns() {
		setChanged();

		if (grid != null)
			grid.notifyPatternsChanged();
	}

	private MatterGrid grid;

	@Override
	public int getSlots() {
		return patterns.getContainerSize();
	}

	@Nonnull
	@Override
	public ItemStack getStackInSlot(int slot) {
		return patterns.getItem(slot);
	}

	@Nonnull
	@Override
	public ItemStack insertItem(int slot, @Nonnull ItemStack stack, boolean simulate) {
		if (stack.getCapability(MatteryCapability.PATTERN).resolve().isEmpty())
			return stack;

		ItemStack self_stack = patterns.getItem(slot);

		if (self_stack.isEmpty()) {
			if (!simulate) {
				patterns.setItem(slot, stack.copy());
			}

			return ItemStack.EMPTY;
		} else if (self_stack.isStackable() && self_stack.getMaxStackSize() > self_stack.getCount() && ItemStack.isSameItemSameTags(self_stack, stack)) {
			int new_count = Math.min(self_stack.getMaxStackSize(), self_stack.getCount() + stack.getCount());
			int diff = new_count - self_stack.getCount();

			if (diff != 0) {
				if (!simulate) {
					self_stack.grow(diff);
					patterns.setChanged();
				}

				ItemStack copy = stack.copy();
				copy.shrink(diff);
				return copy;
			}
		}

		return stack;
	}

	@Nonnull
	@Override
	public ItemStack extractItem(int slot, int amount, boolean simulate) {
		if (amount <= 0)
			return ItemStack.EMPTY;

		ItemStack stack = patterns.getItem(slot);

		if (stack.isEmpty())
			return ItemStack.EMPTY;

		if (!stack.isStackable()) {
			if (!simulate) {
				patterns.setItem(slot, ItemStack.EMPTY);
			}

			return stack.copy();
		}

		ItemStack copy = stack.copy();
		copy.setCount(Math.min(amount, stack.getCount()));

		if (!simulate) {
			stack.shrink(copy.getCount());
			patterns.setChanged();
		}

		return copy;
	}

	@Override
	public CompoundTag save(CompoundTag nbt) {
		nbt.put("patterns", patterns.serializeNBT());
		return super.save(nbt);
	}

	@Override
	public void load(CompoundTag nbt) {
		super.load(nbt);
		patterns = patterns.of(nbt.get("patterns"));
	}

	@Override
	public int getSlotLimit(int slot) {
		return 1;
	}

	@Override
	public boolean isItemValid(int slot, @Nonnull ItemStack stack) {
		return stack.getCapability(MatteryCapability.PATTERN).isPresent();
	}

	@Override
	public void setLevel(Level p_155231_) {
		super.setLevel(p_155231_);

		if (grid == null)
			MatterGrid.scheduleDiscoverNeighbours(this, getBlockPos(), p_155231_);
	}

	@Nullable
	@Override
	public MatterGrid getMatterGrid() {
		return grid;
	}

	@Nullable
	@Override
	public IMatterHandler getMatterHandler() {
		return null;
	}

	@Nullable
	@Override
	public IPatternStorage getPatternStorage() {
		return valid ? this : null;
	}

	private boolean valid = true;

	@Override
	public boolean isValidMatterCell() {
		return valid;
	}

	@Override
	public void setMatterGrid(MatterGrid grid) {
		this.grid = grid;
	}

	@Override
	public void invalidateCaps() {
		super.invalidateCaps();
		valid = false;
	}

	@Override
	public void reviveCaps() {
		super.reviveCaps();
		valid = true;
	}

	private LazyOptional<BlockEntityPatternStorage> resolver = LazyOptional.of(() -> this);

	@Nonnull
	@Override
	public <T> LazyOptional<T> getCapability(@Nonnull Capability<T> cap, @Nullable Direction side) {
		if (cap == MatteryCapability.PATTERN || cap == CapabilityItemHandler.ITEM_HANDLER_CAPABILITY)
			return resolver.cast();

		return super.getCapability(cap, side);
	}

	private static final TranslatableComponent NAME = new TranslatableComponent("block.overdrive_that_matters.pattern_storage");

	@Override
	protected Component getDefaultDisplayName() {
		return NAME;
	}

	@Nullable
	@Override
	public AbstractContainerMenu createMenu(int containerID, Inventory inventory, Player ply) {
		return new PatternStorageMenu(containerID, inventory, this);
	}

	@Override
	public Collection<PatternState> getStoredPatterns() {
		ArrayList<PatternState> list = new ArrayList<>();
		patterns.consumeCapability(MatteryCapability.PATTERN, capability -> list.addAll(capability.getStoredPatterns()));
		return ImmutableList.copyOf(list);
	}

	@Override
	public int getCapacity() {
		AtomicInteger stored = new AtomicInteger();
		patterns.consumeCapability(MatteryCapability.PATTERN, capability -> stored.getAndAdd(capability.getCapacity()));
		return stored.get();
	}

	@Override
	public int getStored() {
		AtomicInteger stored = new AtomicInteger();
		patterns.consumeCapability(MatteryCapability.PATTERN, capability -> stored.getAndAdd(capability.getStored()));
		return stored.get();
	}

	@Override
	public boolean insertPattern(PatternState pattern, boolean only_update, boolean simulate) {
		for (IPatternStorage storage : patterns.capabilityIterator(MatteryCapability.PATTERN)) {
			if (storage.insertPattern(pattern, only_update, simulate)) {
				if (!simulate) {
					setChangedPatterns();
				}

				return true;
			}
		}

		return false;
	}
}
